/**
 * This code starts the VEBus device firmware update.
 *
 * After opening the vff/vfd file (see vff_file.h) call vbdupUpdate to start the update.
 *
 * Currently supported bootloaders:
 * - MK3
 * - P18K (Multi 26/27)
 *
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <velib/mk2/mk2.h>
#include <velib/mk2/hal.h>
#include <velib/utils/ve_todo.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_timer.h>
#include <velib/vebusupd/vff_file.h>
#include <velib/vebusupd/vebus_device_update.h>
#include <velib/vebusupd/mk3_update.h>
#include <velib/vebusupd/pic18k_new_update.h>

/* Connect info defines */
static const char CI_MK2[] = "MK2";
static const char CI_MK3[] = "MK3";
static const char CI_P18K[] = "P18K";
static const char CI_SWL[] = "SWL";
static const char CI_OLD_FASHIONED[] = "OldFashioned";

/* Code part defines */
static const char CP_MAIN[] = CPN_MAIN;
static const char CP_BOOT_BLOCK[] = CPN_REBOOTER;
static const char PN_VERSION[] = CPT_VERSION ;


static VbdupUpdateRxByte* pVebusDeviceRxUpdateRxByte = NULL;

/**
 * This function registers the mk2UpdateRxByte function
 *
 * @todo This should change in the mk2_serial driver, to include pCi.
 * We are now limited to one instance.
 *
 * @param pCi
 */
static void registerMk2UpdateRxByte(VbdupConnectInfo *pCi)
{
	pVebusDeviceRxUpdateRxByte = pCi->pVbdupUpdateRxByte;
	veTodo();
}

/**
 * Retrieve the connect info from the VffData buffer.
 *
 * @param pCi
 */
static void vbdupGetConnectInfo (VbdupConnectInfo* pCi)
{
	int len;
	char buf[15];
	VffData* pData;

	pCi->ct = CT_UNKNOWN;
	pCi->cm = CM_UNKNOWN;

	pData = pCi->pVffData;

	len = vffGetParameterValueAsString(pData, pCi->cpIndex, buf, sizeof(buf), CPT_CONNECT_INFO);
	if (len != 0){

		switch (buf[0]) {
			case 'p':
				pCi->cm = CM_PASSIVE;
				break;
			case 'a':
				pCi->cm = CM_ACTIVE;
				break;
			default:
				break;
		}

		if(memcmp(CI_MK2, &buf[1], sizeof(CI_MK2)) == 0)
			pCi->ct = CT_MK2;
		else if(memcmp(CI_MK3, &buf[1], sizeof(CI_MK3)) == 0)
			pCi->ct = CT_MK3;
		else if(memcmp(CI_P18K, &buf[1], sizeof(CI_P18K)) == 0)
			pCi->ct = CT_PIC18K_NEW;
		else if(memcmp(CI_SWL, &buf[1], sizeof(CI_SWL)) == 0)
			pCi->ct = CT_SWL;
		else if(memcmp(CI_OLD_FASHIONED, &buf[1], sizeof(CI_OLD_FASHIONED)) == 0)
			pCi->ct = CT_OLD_FASHIONED;
	}
}

void vbdupUpdateDone(VbdupUpdateResult result)
{
	logI("vbdup", "Update Done. Result = %d", result);
}

/**
 * Starts update with info from CP_MAIN.
 * After opening the vff/vfd file run this to start the update.
 *
 * @param pCi Connect info
 * @return veFalse when update failed to start
 */
veBool vbdupUpdate (VbdupConnectInfo *pCi)
{
	int i;
#if CFG_WITH_PIC18K_NEW_UPDATE == 1
	int version, model;
#endif
	VffData* pData;

	pCi->pVbdupUpdateBusy=NULL;
	pCi->pVbdupFirmwareUpdateTick=NULL;
	pCi->pVbdupUpdateRxByte = NULL;

	pData = pCi->pVffData;

	i = vffGetIndexOfCodePart(pData, (char*) CP_MAIN);

	if (i == -1) {
		logE("vbdup", "Code part %s not found", CP_MAIN);
		return veFalse;
	}

	pCi->cpIndex = i;

	vbdupGetConnectInfo(pCi);

	switch (pCi->ct) {
		case CT_MK3:
#if CFG_WITH_MK3_UPDATE == 1
			/* Try to start update from MK3 application. If this failes, the update will start from bootloader  */
			logI("vbdup", "Start MK3 firmware update");
			pCi->pVbdupUpdateBusy=mk3UpdateBusy;
			pCi->pVbdupFirmwareUpdateTick=mk3FirmwareTick;
			pCi->pVbdupUpdateRxByte = mk3RxUpdateRxByte;
			registerMk2UpdateRxByte(pCi); /* This should change */

			mk3FirmwareUpdate(START_UPDATE_FROM_APPLICATION, pData, i, vbdupUpdateDone);
#endif
			break;
		case CT_MK2:
#if CFG_WITH_MK2_UPDATE == 1
			veTodo();
#endif
		case CT_PIC18K_NEW:
		case 0:				/* If no connect type is defined in file we assume we want to update Multi firmware */
#if CFG_WITH_PIC18K_NEW_UPDATE == 1
			if(vffGetParameterValueAsInt(pData, i, &version, (char *) PN_VERSION)) {
				/* version layout: mmmmrrr
				 * mmmm = 4 digit model ID
				 * rrr = 3 digit software revision
				 *
				 * supported models with PIC18K new processor > 20xx
				 * 19xx/20xx = old PIC 240V/110V models
				 * 26xx/27xx = new PIC 240V/110V models
				 * */
				model = version / 100000;
				if (model  <= 20) {
					logE("vbdup", "Multi model %d not supported", model);
					return veFalse;
				}
			}
			else {
				logE("vbdup", "No version info found in Main code part");
				return veFalse;
			}
			logI("vbdup", "Start PIC18K_NEW firmware update");
			pCi->pVbdupUpdateBusy=p18knUpdateBusy;
			pCi->pVbdupFirmwareUpdateTick=p18knFirmwareTick;
			pCi->pVbdupUpdateRxByte = p18knRxUpdateRxByte;
			registerMk2UpdateRxByte(pCi); /* This should change */

			p18knFirmwareUpdate(pData, i, vbdupUpdateDone);
#endif
			break;
		case CT_SWL:
		case CT_OLD_FASHIONED:
		default:
			logE("vbdup", "Vebus device update failed, device type %d not supported", pCi->ct);
			return veFalse;
	}

	return veTrue;
}

/**
 * Starts update with info from CP_BOOTBLOCK.
 * After opening the vff/vfd file run this to start the update.
 * @todo
 *
 * @param pCi Connect info
 * @return veFalse when update failed to start
 */
veBool vbdupUpdateBootblock (VbdupConnectInfo *pCi)
{
	veTodo();
	return veFalse;
}

/**
 * Returns veTrue if a VEBus device update is in progress
 *
 * @param pCi
 *
 * @return veTrue if busy
 */
veBool vbdupUpdateBusy(VbdupConnectInfo *pCi)
{
	if (pCi->pVbdupUpdateBusy!= NULL)
		return pCi->pVbdupUpdateBusy();

	return veFalse;
}

/**
 * Firmware update statemachine tick.
 * Called from main every 50ms
 *
 * @param pCi
 */
void vbdubUpdateTick(VbdupConnectInfo *pCi)
{
	if (pCi->pVbdupFirmwareUpdateTick != NULL)
		pCi->pVbdupFirmwareUpdateTick();
}

/**
 * on interrupt! Called from mk2_serial.c
 * @todo See registerMk2UpdateRxByte
 *
 * @param inbyte
 */
void vebusDeviceRxUpdateRxByte(un8 inbyte)
{
	if (pVebusDeviceRxUpdateRxByte != NULL)
		pVebusDeviceRxUpdateRxByte(inbyte);
}
