/**
 * VFF/VFD file parser functions
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <velib/base/ve_string.h>
#include <velib/utils/ve_logger.h>
#include <velib/vebusupd/vff_file.h>

#define min(A,B) ((A)<(B) ? (A) : (B))


/**
 * Decrypt a VFF file string
 *
 * @param pStr pointer to the string
 * @param len String length
 */
static void vffSimpleDecryptString (char* pStr, int len)
{
	int i;

	for (i = 0; i < len; i ++) {
		pStr[i] = pStr[i] - (((i+1) & 0xFF) + 11);
	}
}

/**
 * Read a un16 from buffer rp and return the pointer to the next location
 *
 * @param pVal pointer to resulting value location.
 * @param rp read pointer.
 *
 * @return rp to next
 */
static char* vffLineBufReadUn16 (un16* pVal, char* rp)
{
	* pVal = (un16)(* rp++) & 0xFF;
	* pVal |= ((un16)(* rp++) & 0xFF) << 8;

	return rp;
}

/**
 * @brief Read an un32 from buffer rp and return the pointer to the next location
 *
 * @param pVal	pointer to resulting value location.
 * @param rp	read pointer.
 *
 * @return			rp to next
 */
static char* vffLineBufReadUn32 (un32* pVal, char* rp)
{
	* pVal = (un32)(* rp++) & 0xFF;
	* pVal |= ((un32)(* rp++) & 0xFF) << 8;
	* pVal |= ((un32)(* rp++) & 0xFF) << 16;
	* pVal |= ((un32)(* rp++) & 0xFF) << 24;

	return rp;
}

/**
 * Write a un16 to buffer wp and return the pointer to the next location
 *
 * @param val value to be written.
 * @param wp write pointer.
 *
 * @return wp to next
 */
static char* vffLineBufWriteUn16 (un16 val, char* wp)
{
	* wp++ = (char) (val & 0xFF);
	* wp++ = (char) (val >> 8);

	return wp;
}


/**
 * Search for a given parameter by name.
 * Parameters are ',' separated.
 *
 * example parameter string:
 * "Part=Main,MinBFV=0,ConnectInfo=pMK3,Device=MK3,Version=1170135,Nonce=0005006E,KeyID=2,EL=16,CL=1!"
 *
 * pParameterList strings are Delphi type strings with 2 heading length bytes
 *
 * @param pParameterList	Parameter string (first to bytes are length)
 * @param pParVal			'\0' terminated parameter value string
 * @param parValLen			Number of available characters in the result buffer
 * @param pName				Parameter name
 *
 * @return  Number of characters in ParVal string
 */
static int vffSeekParameter(char* pParameterList, char* pParValue, int parValLen, char* pName)
{
	un16 len, nameLen, clParValLen;
	char* rp;
	char* bep;					/* buffer end pointer */
	char* pep;					/* parameter end pointer */

	rp = pParameterList;

	rp = vffLineBufReadUn16(&len, rp);

	bep = rp + len;
	pep = bep;

	nameLen = strlen(pName);

	if (nameLen > len)
		return 0;				/* Name longer than parameter list */

	while (rp < bep)
	{

		/* Find parameter separation (next comma or end of line) */
		pep = memchr(rp, ',', (bep-rp));
		if (pep == NULL)
			pep = bep;


		if(memcmp(pName, rp, nameLen) == 0) {
			/* String found */
			rp += nameLen;
			if (*rp != '=')
				return 0;		/* Format error */

			rp++;				/* Skip '=' */

			/* Clip the parameter value length to the result buffer length */
			clParValLen = min((pep - rp), (parValLen - 1)); /* -1 for '\0' string termination */

			memcpy(pParValue, rp, clParValLen);
			pParValue += clParValLen;
			*pParValue = '\0';			/* Add string termination */

			return clParValLen;
		}
		rp = pep + 1;			/* Move to next parameter */
	}
	return 0;
}

/**
 * Search for a given parameter by name and return it's value as int.
 *
 * @see vffSeekParameter()
 *
 * Note: returns 0 when parameter value is not a number.
 *
 * @param pValue			pointer to int result
 * @param pParameterList	Parameter string (first to bytes are length)
 * @param pName				Parameter name
 *
 * @return veTrue on success. veFalse when parameter not found
 */
static veBool vffParameterToInt(int* pValue, char* pParameterList, char* pName)
{
	int result;
	char buf[20];

	result = vffSeekParameter(pParameterList, buf, 20, pName);
	if (result == 0)
		return veFalse;

	*pValue = atoi(buf);

	return veTrue;
}


/**
 * Read len bytes from the VFF file.
 *
 * Checksum calculation will be updated on the fly
 *
 * @param pVffFile
 * @param pBuf
 * @param len
 *
 * @return errorCode
 */
static int vffReadBytes(VffFile* pVffFile, char* pBuf, un32 len)
{
	un32 i;
	int result;

	result = fread(pBuf, sizeof(char), len, pVffFile->pFile);
	if (result != len) {
		logE("vffFile","Read Failed, result = %d", result);
		return VFF_ERR_FILE_READ;
	}

	for (i = 0; i < len; i++)
		pVffFile->checksum -= (un8) pBuf[i];

	return VFF_ERR_NONE;
}

/**
 * Get next line.
 *
 * The strings in the file are Delphi type strings.
 * First byte(s) indicate length
 * From VFF_FILE_VERSION > 2: string length defaults 2
 *
 * @param pVffFile
 * @param pData
 * @param stringIsEncrypted Decrypt string
 * @param use1ByteAsLen If TRUE string length is one byte.
 * @param stringIsCtype false is Delphi strings, true is C strings
 * *
 * @return errorCode
 */
static int vffGetNextLine (VffFile* pVffFile, VffData* pData, veBool stringIsEncrypted, veBool use1ByteAsLen, veBool stringIsCtype)
{
	char  csCalc, csFile;
	char* pStr;
	char* pStrB;
	un16 nLen, len, i;
	int result;


	pStr = pData->pWrite;
	nLen = 1;					/* number of length bytes */

	if (!use1ByteAsLen)
		nLen ++;

	/* Get string length */
	result = vffReadBytes(pVffFile, pStr, nLen);
	if (result != VFF_ERR_NONE)
		return result;

	len =  (un16) pStr[0] & 0xFF;
	if (!use1ByteAsLen)
		len |= ((un16) pStr[1] & 0xFF) << 8;

	pStrB = pStr;

	if (!stringIsCtype)
		pStrB = vffLineBufWriteUn16(len, pStrB); /* Always use 2 bytes for len in line buffer */

	result = vffReadBytes(pVffFile, pStrB, len);
	if (result != VFF_ERR_NONE)
		return result;

	if (stringIsEncrypted)
		vffSimpleDecryptString(pStrB, len);

	/* Check string checksum */
	csCalc = 0;
	for (i = 0; i < len; i++)
		csCalc += (un8) pStrB[i];

	result = vffReadBytes(pVffFile, &csFile, 1);
	if (result != VFF_ERR_NONE)
		return result;

	if (csFile != csCalc) {
		logE("vffFile","File format line checkum error");
		return VFF_ERR_FILE_FORMAT;
	}

	if (stringIsCtype) {
		pData->pWrite += len;			/* Update the write pointer */
		*pData->pWrite++ = '\n';
		*pData->pWrite++ = '\0';
	}
	else
		pData->pWrite += len+2;			/* Update the write pointer */

	return VFF_ERR_NONE;
}


/**
 * Read Code lines. The number of lines are defined in de parameter list.
 *
 * @param pVffFile
 * @param pData
 * @param codePartIndex
 * @param name The name of the code section
 *
 * @return errorCode
 */
static int readCodeLines(VffFile* pVffFile, VffData* pData, int codePartIndex, char* name)
{
	int nLines, i, result;

	i = codePartIndex;

	if(!vffParameterToInt(&nLines, pData->codePart[i].pParameterLines, name))
		nLines = 0;				/* No Code lines in file */

	if (nLines == 0){
		/* Add 00 length bytes to line buffer*/
		*pData->pWrite++ = 0;
		*pData->pWrite++ = 0;
	}

	for (i = 0; i < nLines; i ++) {
		result = vffGetNextLine(pVffFile, pData, veFalse, veFalse, veFalse); /* read un-encrypted, 2 byte length, Delphi string */
		if (result != VFF_ERR_NONE)
			return result;
	}

	return VFF_ERR_NONE;
}

/**
 * Read Message lines. The number of lines are defined in de parameter list.
 * Message lines are merged in one C type string with '\n' line separators.
 *
 * @param pVffFile
 * @param pData
 * @param codePartIndex
 * @param name The name of the message section
 *
 * @return errorCode
 */
static int readMessageLines(VffFile* pVffFile, VffData* pData, int codePartIndex, char* name)
{
	int nLines, i, result;

	i = codePartIndex;

	if(!vffParameterToInt(&nLines, pData->codePart[i].pParameterLines, name))
		nLines = 0;				/* No Code lines in file */

	if (nLines == 0){
		*pData->pWrite++ = '\0';
	}

	for (i = 0; i < nLines; i ++) {
		result = vffGetNextLine(pVffFile, pData, veTrue, veFalse, veTrue); /* read encrypted, 2 byte length, C string */
		if (result != VFF_ERR_NONE)
			return result;
		pData->pWrite--; /* discard '\0' in order to append new strings */
	}
	*pData->pWrite++ = '\0'; /* Close C string */

	return VFF_ERR_NONE;
}


/**
 * Init the VffFile struct and rewind pFile
 *
 * @param pFile
 * @param pVffFile
 */
void vffFileInit (FILE* pFile, VffFile* pVffFile)
{
	pVffFile->pFile = pFile;

	pVffFile->checksum = 0;

	fseek(pFile, 0, SEEK_END); /* end of file */
	pVffFile->endLocation = ftell(pFile);
	fseek(pFile, 0, SEEK_SET); /* beginning of file */
}


/**
 * Open and init a VFF file by name
 *
 * @param pVffFile
 * @param fileName
 *
 * @return errorCode
 */
int vffFileOpen (VffFile* pVffFile, char* fileName)
{
	FILE *pFile;

	logI("vffFile", "Open %s for reading", fileName);

	if (!(pFile = fopen(fileName, "rb")))
	{
		logE("vffFile","File open failed");
		return VFF_ERR_FILE_OPEN;
	}

	vffFileInit(pFile, pVffFile);

	return VFF_ERR_NONE;
}

/**
 * Close the VFF file
 *
 * @param pVffFile
 */
void vffFileClose (VffFile* pVffFile)
{
	fclose(pVffFile->pFile);
}

/**
 * Malloc a VFF data buffer
 *
 * A data buffer can contain multiple code parts.
 * The size can be derived from the VFF file endIndex
 * The contents will be a bit smaller. There are no file and line checksums.
 * Since we don't know the amount of lines yet we can't tell how much smaller
 * the buffer can be.
 *
 * @param pData
 * @param size
 *
 * @return errorCode
 */
int vffMallocDataBuffer(VffData* pData, int size)
{
	int i;

	for (i = 0; i < VFF_N_CODE_PARTS; i++){
		pData->codePart[i].pMessageLines = NULL;
		pData->codePart[i].pEEPROMLines = NULL;
		pData->codePart[i].pCodeLines = NULL;
		pData->codePart[i].pParameterLines = NULL;
	}

	/* Malloc a line buffer */
	pData->pStart = malloc(size);
	if (pData->pStart == NULL) {
		logE("vffFile", "Malloc VFF line buffer failed");
		return VFF_ERR_MALLOC;
	}
	pData->size = size;
	pData->pWrite = pData->pStart;
	pData->pRead = pData->pStart;
	return VFF_ERR_NONE;
}

/**
 * Free the line buffer
 *
 * @param pData
 */
void vffFreeDataBuffer(VffData* pData)
{
	/* Free line buffer allocated memory */
	free(pData->pStart);
	pData->size = 0;
}

/**
 * Read the file, decrypt applicable lines and add data to VffCodePart.
 * The EEPROM lines and Code lines remain encrypted.
 * There can be multiple code parts within the file.
 *
 * @param pVffFile
 * @param pData
 *
 * @return errorCode
 */
int vffParseFile(VffFile* pVffFile, VffData* pData)
{
	int result, i;
	un32 fileChecksum;
	char buf[4];

	pData->pWrite = pData->pStart;

	for (i = 0; i < VFF_N_CODE_PARTS; i ++) {
		logI("vffFile", "Parsing code part %d", i);

		pVffFile->checksum = 0;

		/* Get file format ID */
		pData->codePart[i].pFileFormatId = pData->pWrite;
		result = vffGetNextLine(pVffFile, pData, veFalse, veTrue, veFalse); /* read unencrypted, 1 byte length, Delphi string */
		if (result != VFF_ERR_NONE)
			return result;

		/* Check fileformat ID */
		pData->codePart[i].pFileFormatId += 2;			/* Skip length bytes */

		if(strncmp(VFF_FILE_ID_STR, pData->codePart[i].pFileFormatId, 28) != 0){
			/* File format ID does not match */
			logE("vffFile","File format error");
			return VFF_ERR_FILE_FORMAT;
		}

		/* read 2bytes file version info */
		result = vffReadBytes(pVffFile, buf, 2);
		if (result != VFF_ERR_NONE)
			return result;

		vffLineBufReadUn16(&pData->codePart[i].fileVersion, buf);
		if (pData->codePart[i].fileVersion != VFF_FILE_VERSION) {
			logE("vffFile","File version mismatch. file = %d, expected = %d", pData->codePart[i].fileVersion, VFF_FILE_VERSION);
			return VFF_ERR_FILE_VERSION;
		}

		result = fread(buf, sizeof(char), 4, pVffFile->pFile); /* RAW read 4 bytes file checksum without updating checksum calculation */
		if (result != 4)
			return VFF_ERR_FILE_READ;

		vffLineBufReadUn32(&fileChecksum, buf);
		pVffFile->checksum += fileChecksum;

		/* Get a comma separated parameter list */
		pData->codePart[i].pParameterLines = pData->pWrite;
		result = vffGetNextLine(pVffFile, pData, veTrue, veFalse, veFalse); /* read encrypted, 2 byte length, Delphi string */
		if (result != VFF_ERR_NONE)
			return result;

		/* Get calibration map */
		pData->codePart[i].pCalibrationMap = pData->pWrite;
		result = vffGetNextLine(pVffFile, pData, veTrue, veFalse, veFalse); /* read encrypted, 2 byte length, Delphi string */
		if (result != VFF_ERR_NONE)
			return result;

		/* Read message,eeprom and code lines. The order of this is fixed */
		pData->codePart[i].pMessageLines = pData->pWrite;
		result = readMessageLines(pVffFile, pData, i, CPT_MESSAGE);
		if (result != VFF_ERR_NONE)
			return result;

		pData->codePart[i].pEEPROMLines = pData->pWrite;
		result = readCodeLines(pVffFile, pData, i, CPT_EEPROM);
		if (result != VFF_ERR_NONE)
			return result;

		pData->codePart[i].pCodeLines = pData->pWrite;
		result = readCodeLines(pVffFile, pData, i, CPT_CODE);
		if (result != VFF_ERR_NONE)
			return result;

		/* Verify checksum */
		if(pVffFile->checksum != 0)
			return VFF_ERR_FILE_CHECKSUM;

		pData->nCodeParts = i + 1 ;	/* Code part index complete */

		if (ftell(pVffFile->pFile) >= pVffFile->endLocation)
			break;				/* EOF no more code parts.*/

	}

	return VFF_ERR_NONE;
}


/**
 * Find the index of a code part by name
 *
 * @param pData
 * @param name
 *
 * @return index. -1 if not found
 */
int vffGetIndexOfCodePart(VffData* pData, char* pName)
{
	int i, len, nameLen;
	char buf[20];

	nameLen = strlen(pName);

	if (nameLen > sizeof(buf))
		return -1;

	for (i = 0; i < pData->nCodeParts; i ++) {
		len = vffSeekParameter(pData->codePart[i].pParameterLines, buf, 20, CPT_PART);
		if (len > 0) {

			if(memcmp(pName, buf, nameLen) == 0)
				return i;		/* Found codepart */
		}
	}

	return -1;
}

/**
 * Search for a given parameter by name.
 * Parameters are ',' separated.
 *
 * example parameter string:
 * "Part=Main,MinBFV=0,ConnectInfo=pMK3,Device=MK3,Version=1170135,Nonce=0005006E,KeyID=2,EL=16,CL=1!"
 *
 * pParameterList strings are Delphi type strings with 2 heading length bytes
 *
 * @param pData				vddData struct
 * @param cpIndex			codepart index
 * @param pParVal			'\0' terminated parameter value string
 * @param parValLen			Number of available characters in the result buffer
 * @param pName				Parameter name
 *
 * @return  Number of characters in ParVal string
 */
int vffGetParameterValueAsString(VffData* pData, int cpIndex, char* pParValue, int parValLen, char* pName)
{
	return vffSeekParameter(pData->codePart[cpIndex].pParameterLines, pParValue, parValLen, pName);
}


/**
 * Search for a given parameter by name and return it's value as int.
 *
 * @see vffGetParameterValueAsString()
 *
 * Note: returns 0 when parameter value is not a number.
 *
 * @param pData
 * @param cpIndex
 * @param pValue			pointer to int result
 * @param pName				Parameter name
 *
 * @return veTrue on success. veFalse when parameter not found
 */
veBool vffGetParameterValueAsInt(VffData* pData, int cpIndex, int* pValue, char* pName)
{
	return vffParameterToInt(pValue, pData->codePart[cpIndex].pParameterLines, pName);
}

/**
 * return the line length in pVal and return the start of data address.
 *
 * @param pVal
 * @param rp
 *
 * @return updated buffer read pointer
 */
char* vffGetLineLength (un16* pVal, char* rp)
{
	return vffLineBufReadUn16 (pVal, rp);
}


/**
 * Convert the eeprom line buffer to raw eeprom buffer.
 * The eeprom raw buffer contains raw eeprom values only.
 * No Len, HID, TID and location info
 *
 * @param eepromLines
 * @param eepromRaw
 */
void vffEepromLineBufToRaw(char* eepromLines, char* eepromRaw)
{
	int i;
	char *bp;

	bp = eepromLines;

	for (i = 0; i < 4; i ++) {
		bp += 2 + 2 + 4;	/* skip len, HID and TID */
		memcpy(& eepromRaw[i * 64], bp, 64);
		bp += (64 + 3);			/* skip data and 3 location bytes */
	 }
}

/**
 * Merge the raw eeprom buffer back into the linebuffer.
 *
 * @param eepromRaw
 * @param eepromLines
 */
void vffMergeEepromRawInLineBuf(char* eepromRaw, char* eepromLines)
{
	int i;
	char *bp;

	bp = eepromLines;

	for (i = 0; i < 4; i ++) {
		bp += 2 + 2 + 4;	/* skip len, HID and TID */
		memcpy(bp, & eepromRaw[i * 64], 64);
		bp += (64 + 3);			/* skip data and 3 location bytes */
	 }
}

/**
 * debug function to print contents of data buffer
 *
 * @param pData VffCodePart
 */
void vffDataBufferToLog(VffData* pData)
{
	un16 len;
	char* rp;
	int i, nLines;

	logI("vffFile", "---------------------------------------------------------------");
	logI("vffFile", "VFF line buffer dump.");
	logI("vffFile", "%d code parts", pData->nCodeParts);


	for (i = 0; i < pData->nCodeParts; i++)
	{
		logI("vffFile", "--------------------------");
		logI("vffFile", "code part %d", i);
		logI("vffFile", "VFF file version = %d", pData->codePart[i].fileVersion);
		logI("vffFile", "Parameter lines");

		rp = pData->codePart[i].pParameterLines;
		rp = vffLineBufReadUn16(&len, rp);
		logI("vffFile", "len = %d |%.*s|", len, len, rp);

		nLines = 0;
		vffGetParameterValueAsInt(pData, i, & nLines, CPT_MESSAGE);

		if(nLines > 0) {
			rp = pData->codePart[i].pMessageLines;
			logI("vffFile","Message dump\n %s", rp);
		}
	}
}
