/* MK3 selftest. (Production test)
 *
 *
 * The test setup contains a tester MK3 and the DUT MK3.
 * The VEbus should be powered using a test box, or a Multi in freeze mode.
 * (We don't want Multi vebus communication during testing)
 * Communication between them is performed using the panel frame format.
 * Both MK3's should be configured to forward panel frames.
 * The panel frame contents is non standard and for testing only.
 *
 * Panel frame format for testing
 *
 * 0x40, testnr, pdet_stdby_line_flags, xx, xx
 *
 * pdet_stdby_line_flags:
 *   -bit 0 = panel detect
 *   -bit 1 = standby line level
 *
 * The Tester should return the standby line and panel detect line levels.
 *
 * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WINDOWS
#include <conio.h>
#include <sys/stat.h>
#endif

#include <velib/velib_inc_mk2.h>

#include <velib/base/base.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2_update.h>
#include <velib/mk2/vebus_device.h>
#include <velib/mk3/mk3.h>
#include <velib/platform/console.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_timer.h>


#include <../../inc/mk3_firmware_upgrade.h> /* use the same firmware version as mk2-dbus */

typedef void DelayedExecCallBack(void);

typedef enum {
	ST_RESULT_OK,
	ST_RESULT_FAIL_BUS_FREE_LINE,
	ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE,
	ST_RESULT_FAIL_VEBUS_TX,
	ST_RESULT_FAIL_STANDBY_LINE,
	ST_RESULT_FAIL_UNKNOWN_CAUSE, /* Up to here we are compatible with mk2selftest */
	ST_RESULT_FAIL_PANELDETECT_LINE
} SelfTestResult;

static un8 delayedExecTimer;
static DelayedExecCallBack* delayedExecCb;
static un8 startupTimer;
struct Mk2S mk2;
static veBool testerMode = veFalse; /* if veTrue we act as tester. default mode = DUT */
static int testNr = 0;
static veBool pollUserInput = veFalse;
static SelfTestResult testResult;

extern int skipHwLineTests;
extern int setStandbyLine;
extern int setPanelDetectLine;
extern int skipUpdate;

/* Function declarations */
static void setStandbyAndPanelDetectLines(Mk2Callback* callback);
static void settingLinesDone(void);
static void mk3UpdateDone(Mk2Result result);

/* This function swaps the high and low nible of a byte */
static un8 nibSwap(un8 x)
{
	return ((x & 0xF) << 4) | (x >> 4);
}

/* Calculate the vebus frame checksum */
static sn8 vebusFrCalcChecksum(un8* data, un8 len)
{
	un8 i;
	sn8 cs;

	cs = 0;

	for (i = 0; i < 5; i=i+2)
		cs -= nibSwap(data[i]);

	for (i = 0; i < len; i++)
		cs -= data[i];

	cs -= len;

	return cs;
}


static void vebusfrSendTestFrame (un8 testNr, un8 lineFlags, Mk2Handler* handler, Mk2Callback* callback)
{
	Mk2MsgTx* msg;

	msg = mk2TxAlloc();
	if (!msg)
		return;

	msg->addChecksum = veFalse;

	msg->buf[0] = 0x40; /* Panel frame ID */
	msg->buf[1] = testNr;
	msg->buf[2] = lineFlags;
	msg->buf[3] = 0x55;			/* Don't care data */
	msg->buf[4] = 0xaa;
	msg->buf[5] = vebusFrCalcChecksum(&msg->buf[0], 5);

	msg->length = 5;			/* data length without checksum  */

	msg->handler = handler;
	msg->callback = callback;

	mk2FhQueueFrame(msg);
}

/* Delayed execution functionality through taskTick */
void mk3TestDelayedExec(DelayedExecCallBack* cb, un8 n50msCount)
{
	delayedExecCb = cb;
	delayedExecTimer = n50msCount;
}

void mk2FhRawFrameReceivedEvent(Mk2MsgRx *mk2msg)
{
	mk2Dump("<", mk2msg);
}

void mk2FhFrameSendEvent(Mk2MsgTx *mk2msg)
{
	//mk2Dump(">", (Mk2MsgRx *) mk2msg);
}

void pollPanelAndStanbyLineFlagsUserInput(void)
{
#ifndef WINDOWS
	fd_set fdSelect;
	int retVal;
	struct timeval tv;
#endif
	char c, cf;
	un8 flags;

#ifdef WINDOWS
	if(kbhit())
		c = getchar();
#else
	FD_ZERO(&fdSelect);			/* Init select file descriptor */
	FD_SET(0, &fdSelect);		/* Choose stdin  */

	tv.tv_sec = 0;
	tv.tv_usec = 20;		/* Wait 20us */

	retVal = select(1, &fdSelect, NULL, NULL, &tv);

	if (retVal > 0) {
		c = getchar();
	}
#endif
	else
		return;

	while((cf = getchar()) != '\n' && cf != EOF); /* flush the input buffer */

	switch(c)
	{
		case '0':
			flags = 0;
			break;
		case '1':
			flags = 1;
			break;
		case '2':
			flags = 2;
			break;
		case '3':
			flags = 3;
			break;
		default:
			fprintf(stdout,"Wrong input value. Try again\n");
			return;
	}

	pollUserInput = veFalse;

	vebusfrSendTestFrame(testNr,flags, NULL, NULL);
}

void getPanelAndStanbyLineFlags(int n)
{
	testNr = n;

	if (!pollUserInput)
		fprintf(stdout,"Enter standby and panel detect line state (0..3) (bit 1 = Standby, bit 0 = Panel) \r\n");
	/* Schedule user input check */
	pollUserInput = veTrue;
}

static void mk3UpdateDone(Mk2Result result)
{
	VE_UNUSED(result);

	startupTimer = 20;
}


static void doExit(int ret)
{
	pltExit(ret);
}

static void done(void)
{
	fprintf(stdout,"OK\r\n");
	doExit(0);
}

static void errDone(void)
{
	doExit(1);
}

static void error(SelfTestResult result, un32 additionalInfo)
{
	switch (result)
	{
		case ST_RESULT_OK:
			done();
			break;
		case ST_RESULT_FAIL_BUS_FREE_LINE:
			fprintf(stderr, "ERROR: vebus line free %u\n", additionalInfo);
			break;
		case ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE:
			fprintf(stderr, "ERROR: driver enable %u\n", additionalInfo);
			break;
		case ST_RESULT_FAIL_VEBUS_TX:
			fprintf(stderr, "ERROR: vebus tx %u\n", additionalInfo);
			if (additionalInfo & 4)	{
				fprintf(stderr, "Tester not connected or running?\n");
				errDone();
			}
			break;
		case ST_RESULT_FAIL_UNKNOWN_CAUSE:
			fprintf(stderr, "ERROR: unknown %u\n", additionalInfo);
			break;
		case ST_RESULT_FAIL_STANDBY_LINE:
			fprintf(stderr, "ERROR: standby line %u\n", additionalInfo);
			break;
		case ST_RESULT_FAIL_PANELDETECT_LINE:
			fprintf(stderr, "ERROR: panel detect line %u\n", additionalInfo);
			break;
	}

	testResult = result;
	vebusfrSendTestFrame(5, 0xC0, NULL, NULL);
	mk3TestDelayedExec(errDone, 2); /* Wait 200ms */
}

void mk2FhFrameReceivedEvent(Mk2MsgRx *mk2msg)
{
	un8 testNr;
	un8 flags;

	if (testerMode) {
		if (mk2msg->buf[0] != 0x40 || mk2msg->length < 4)
		return;

		logI("mk3test", "received panel frame");
		/* Handle tester response to panel frames */
		testNr = mk2msg->buf[1];
		flags = mk2msg->buf[2];

		switch (testNr)
		{
			case 1:
				vebusfrSendTestFrame(1,0, NULL, NULL);
				break;
			case 2:
			case 3:
			case 4:
				if (flags == 0)
					getPanelAndStanbyLineFlags(testNr);
				vebusfrSendTestFrame(testNr, 0x80, NULL, NULL);
				break;
			case 5:
				done();

			default:
				break;
		}

		return;
	}

	VE_UNUSED(mk2msg);
}

/*
 * The tester sents the panel frame message back for the selftest,
 * but the mk3 will pass it on. Parse it and check the standby line.
 */
static Mk2EventReply testReplyMessageHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	un8 testNr;
	un8 flags;					/* Stanby line and panel detect flags */

	logI("mk3test", "Test reply message handler");


	if (mk2event == MK2_FRAME && msg->buf[0] == 0x40 && msg->length == 5)
	{
		testNr = msg->buf[1];
		flags = msg->buf[2];

		if (flags == 0x80) {
			/* Waiting for user/tester line level feedback */
			logI("mk3test", "Waiting for user/tester feedback");
			pollUserInput = veTrue;
			return MK2_DONE;
		}

		pollUserInput = veFalse;

		switch (testNr)
		{
			case 2:
				/* standby & panel detect line should be high */
				if (flags != 3) {
					if ((flags & 1) == 0) {
						error(ST_RESULT_FAIL_PANELDETECT_LINE, 0);
						return MK2_DONE;
					}
					if ((flags & 2) == 0) {
						error(ST_RESULT_FAIL_STANDBY_LINE, 0);
						return MK2_DONE;
					}
				}
				else
					logI("mk3test", "Test OK. panel released / standy released");
				return MK2_DONE;
				break;
			case 3:
				/* panel detect should be low, standby high */
				if (flags != 2) {
					if ((flags & 1) != 0) {
						error(ST_RESULT_FAIL_PANELDETECT_LINE, 1);
						return MK2_DONE;
					}
					if ((flags & 2) == 0) {
						error(ST_RESULT_FAIL_STANDBY_LINE, 0);
						return MK2_DONE;
					}
				}
				else
					logI("mk3test", "Test OK. panel detect active / standy released");
				return MK2_DONE;
				break;
			case 4:
				/* both lines should be low  */
				if (flags != 0) {
					if ((flags & 1) != 0) {
						error(ST_RESULT_FAIL_PANELDETECT_LINE, 1);
						return MK2_DONE;
					}
					if ((flags & 2) != 0) {
						error(ST_RESULT_FAIL_STANDBY_LINE, 1);
						return MK2_DONE;
					}
				}
				else
					logI("mk3test", "Test OK. panel detect active / standy active");
				return MK2_DONE;
				break;
			default:
				/* Unkown test number */
				break;
		}
	}
	return MK2_CONTINUE;
}


static void sendLineTestFrameDone(void)
{
	if (testResult != ST_RESULT_OK)
		return;

	if (pollUserInput) {
		/* Ping pong 0x80 <-> 0xC0 until user input done */
		vebusfrSendTestFrame(testNr, 0xC0, testReplyMessageHandler, sendLineTestFrameDone);
		return;
	}

	if (testNr < 4) {
		/* Start next line test */
		testNr++;
		setStandbyAndPanelDetectLines(settingLinesDone);
	}
	else {
		vebusfrSendTestFrame(5, 0xC0, NULL, NULL);
		mk3TestDelayedExec(done, 5); /* Wait 500ms */
	}


}

static void waitForLinesToSettleDone(void)
{
	vebusfrSendTestFrame(testNr, 0, testReplyMessageHandler, sendLineTestFrameDone);
}

static void settingLinesDone(void)
{
	mk3TestDelayedExec(waitForLinesToSettleDone, 2); /* Wait 100ms */
}


static void setStandbyAndPanelDetectLines(Mk2Callback* callback)
{
	switch (testNr)
	{
		case 2:
			fprintf(stdout,"Setting panel released/ standy released\n");
			mk3SetHardwareLines (callback, veFalse, veFalse);
			break;
		case 3:
			fprintf(stdout,"Setting panel / standy released\n");
			mk3SetHardwareLines (callback, veTrue, veFalse);
			break;
		case 4:
			fprintf(stdout,"Setting panel and standy\n");
			mk3SetHardwareLines (callback, veTrue, veTrue);
			break;
		default:
			fprintf(stderr, "Unsupported test number\n");
			doExit(1);
	}

	logI("mk3test", "checking standby released/ panel detect released ");
}

static void mk3SelfTestDone(void)
{
	logI("mk3test", "mk3 self test done");

	if (skipHwLineTests) {
		vebusfrSendTestFrame(5, 0xC0, NULL, NULL);
		mk3TestDelayedExec(done, 5); /* Wait 500ms */
		return;
	}
	testNr = 2;
	setStandbyAndPanelDetectLines(settingLinesDone);
}


/* The expected values to be returned by the MK3 after a self-test */
#define MK3_TEST_RESULT0_MIN					80
#define MK3_TEST_RESULT0_MAX					119

#define MK3_TEST_RESULT1_MASK_RS485_DISABLE		0x01
#define MK3_TEST_RESULT1_MASK_TX_ERROR			0x06

/* parse the selftest response */
static Mk2EventReply selftestHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'T' && msg->length >= 4)
	{
		un8 result0 = msg->buf[2];
		un8 result1 = msg->buf[3];

		fprintf(stdout, "MK3 selftest busFree rise time = %d us\n", result0);

		/* The rise time of the bus free line is outside the permitted range */
		if (result0 < MK3_TEST_RESULT0_MIN || result0 > MK3_TEST_RESULT0_MAX) {
			error(ST_RESULT_FAIL_BUS_FREE_LINE, result0);
			return MK2_DONE;
		}

		fprintf(stdout,"MK3 txEnable testflags = 0x%x\n", result1);

		/* The RS485 driver could not be disabled */
		if (result1 & MK3_TEST_RESULT1_MASK_RS485_DISABLE) {
			error(ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE, result1);
			return MK2_DONE;
		}

		/* There was a problem driving either a hard or a soft 1 */
		if (result1 & MK3_TEST_RESULT1_MASK_TX_ERROR) {
			error(ST_RESULT_FAIL_VEBUS_TX, result1);
			return MK2_DONE;
		}
		/*
		 * The MK2 returned a non-zero result that did not correspond to any
		 * of the known error codes.
		 */
		if (result1) {
			error(ST_RESULT_FAIL_UNKNOWN_CAUSE, result1);
			return MK2_DONE;
		}

		return MK2_DONE;
	}

	if (mk2event == MK2_TIMEOUT) {
		fprintf(stderr, "ERROR: no reply to selftest command\n");
		doExit(1);
	}

	return MK2_CONTINUE;
}

veBool startMk3SelfTest(Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('W');
	if (!msg) {
		veAssert(veFalse);
		return veFalse;
	}

	testResult = ST_RESULT_OK;

	/* Send panel frame. */
	logI("selftest", "sending dummy frame to tester");
	vebusfrSendTestFrame (1, 0, NULL, NULL);

	/* queue the selftest request for the mk2 */
	msg = mk2Msg('T');
	if (!msg) {
		veAssert(veFalse);
		return veFalse;
	}
	msg->handler = selftestHandler;
	msg->callback = callback;
	mk2FhQueueFrame(msg);

	return veTrue;
}

static void setHardwareIdDone(void)
{
	startMk3SelfTest(mk3SelfTestDone);
}

static Mk2EventReply setHardwareIdHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'I' && msg->length > 4)
	{
		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

veBool setHardwareId(Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('I');

	msg->length = 4;
	msg->buf[2] = 0x01; /* first 0x01 indicates to only write hardware id, the value follows */
	msg->buf[3] = 117;
	msg->handler = setHardwareIdHandler;
	msg->callback = callback;
	mk2FhQueueFrame(msg);

	return veTrue;
}


/* 1< mk2 is there, set its state or update when outdated */
static void mk3GetVersionDone(void)
{
	un32 version;
	un32 product;

	version = mk2.version % 1000;
	product = mk2.version - version;

	logI("mk3test","product ID = %d (/1000)", product);

	/* MK3 returns the product ID in decimal. products.h defines it as 0x1170 */
	if (product != (1170 * 1000)) {
		fprintf(stderr, "Product not an MK3");
		doExit(0);				/* Add error code */
	}

	if (version != MK3_UPDATE_VERSION) {
		if (skipUpdate) {
			fprintf(stdout, "wrong mk3 firmware version (%d), expected version %d\n", mk2.version % 1000, MK3_UPDATE_VERSION);
			fprintf(stdout, "skipping firmware update\n");
		}
		else {
			fprintf(stdout, "wrong mk3 firmware version (%d), starting update to version %d\n", mk2.version % 1000, MK3_UPDATE_VERSION);
			mk3FirmwareUpgrade(mk3UpdateDone);
			return;
		}
	}

	if (testerMode){
		fprintf(stdout, "Running in tester mode\n");
		return;
	}

	fprintf(stdout, "Running in DUT mode\n");

	if (setStandbyLine || setPanelDetectLine) {
		mk3SetHardwareLines (done, setPanelDetectLine, setStandbyLine);
		return;
	}
	setHardwareId(setHardwareIdDone);
}


/*
 * This is called early, there is nothing know yet about dev.
 * (besides its short address)
 */
void taskInit(void)
{
	setvbuf(stdout, NULL, _IONBF, 0); /* no buffering  */

	mk2FhInit();
	mk2Init();

	/*
	 * Setup the hardware, these are two distinct calls since the bpp3 does
	 * use the regular pc serial interface, but has hardware to control some
	 * mk2 io pins.
	 */
	mk2HalInit();
	mk2IoInit();

	startupTimer = 10;

	pltInterruptEnable();
}

void taskUpdate(void)
{
	mk2FhUpdate();
}


static void mk3SendSetStateDone(void)
{
	mk2.version = 0;
	logI("mk3test", "getversion");
	mk2GetVersion(&mk2.version, mk3GetVersionDone);
}

/* 1> first get the version of the mk2 */
static void mk2Begin(void)
{
	logI("mk3test", "Configure panel frame forwarding");

	mk2.state.sFlags = MK2_S_FLAG_AUTO_FORWARD_PANEL_FRAMES;
	mk2SetState(&mk2.state, mk3SendSetStateDone);
}

void taskTick(void)
{
	if (mk2UpdateBusy()) {
		mk2FirmwareTick();
		return;
	}

	mk2FhTick();

	if (mk3FirmwareUpgradeBusy()) {
		mk3FirmwareUpgradeTick();
		return;
	}


	if (delayedExecTimer != 0 && --delayedExecTimer==0) {
		if (delayedExecCb != NULL)
			delayedExecCb();
	}

	/*
	 * Get the mk2 out of reset before the startupTimer expires, so it has
	 * some time to startup.
	 */
	if (startupTimer == 5)
		mk2HoldInReset(veFalse);

	/* begin the startup procedure */
	if (startupTimer != 0 && --startupTimer==0)
		mk2Begin();

	if (testerMode && pollUserInput)
		pollPanelAndStanbyLineFlagsUserInput();
}

/** No response is received after several tries. */
void mk2NoResponseCallback(void)
{
	logI("mk3test", "timeout");

	fprintf(stderr, "ERROR: mk3 did not respond\n");
	pltExit(1);
}

veBool vebusCanBeControlled(void)
{
	return veTrue;
}

void mk3SelftestSetTesterMode(veBool flag)
{
	testerMode = flag;
}
